#include "scalar.h"
#include "state.h"

using namespace Frost;

const s_str Scalar::TYPE = "Scalar";
s_double Scalar::epsilon = std::numeric_limits<double>::epsilon();

s_double factorial(s_double d)
{
    if (fabs(d.Get() - 1.0) < 0.1)
        return 1.0;
    else
        return round(d.Get())*factorial(d - 1.0);
}

s_str Scalar::PrintContent() const
{
    s_str sv;
    if (fabs(vi) < epsilon)
        sv = s_str::Convert(vr);
    else if (fabs(vr) < epsilon)
        sv = (vi == 1.0 ? "i" : (vi == -1.0 ? "-i" : vi + " i"));
    else
        sv = vr + (vi < 0.0 ? " - " + (vi == -1.0 ? "i" : (-vi) + " i") : " + " + (vi == 1.0 ? "i" : vi + " i"));

    if (name.IsEmpty())
        return sv;
    else
        return name + " = " + sv;
}

s_str Scalar::GetType() const
{
    return TYPE;
}

s_refptr<Symbol> Scalar::Clone() const
{
    return s_refptr<Symbol>(new Scalar(*this));
}

s_bool Scalar::IsNumber(const s_str& s)
{
    s_str t = s;
    t.Trim(' ');
    if (t.StartsWith("+") || t.StartsWith("-"))
        t.EraseFromStart(1);

    s_bool bCommaFound;
    s_str::iterator iter = t.Begin();
    while (iter != t.End())
    {
        if (*iter == '.')
        {
            if (!bCommaFound) bCommaFound = true;
            else return false;
        }
        else if (!s_str::IsNumber(*iter))
            break;

        ++iter;
    }

    if (iter == t.End())
        return true;

    if (*iter != 'e')
        return false;

    ++iter;
    if (iter == t.End())
        return false;

    if (*iter == '+' || *iter == '-')
    {
        ++iter;
        if (iter == t.End())
            return false;
    }

    while (iter != t.End())
    {
        if (!s_str::IsNumber(*iter))
            return false;

        ++iter;
    }

    return true;
}

s_bool Scalar::MatchesConstructor(const s_str& s)
{
    return IsNumber(s);
}

s_refptr<Symbol> Scalar::CreateFrom(State& mState, const s_str& n, const s_str& s, s_bool inl, s_str& e)
{
    return s_refptr<Symbol>(new Scalar(n, s_double(s)));
}

void Scalar::Register(State& mState)
{
    mState.RegisterSymbolType<Scalar>();
    mState.RegisterConstructor<Scalar>(&Scalar::MatchesConstructor, &Scalar::CreateFrom);

    // Operators
    mState.RegisterLeftUnaryOperator<Scalar>("!", [] (State& s, s_refptr<Symbol> a, s_str& e) -> s_refptr<Symbol> {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(a);
        if (sa->vi != 0.0 || sa->vr < 0.0)
        {
            if (!e.IsEmpty()) e.PushBack('\n');
            e += "\tError : The factorial '!' is only defined for real positive numbers";
            return nullptr;
        }
        return s_refptr<Symbol>(new Scalar("", factorial(sa->vr)));
    });
    mState.RegisterBinaryOperator<Scalar,Scalar>("*", [] (State& s, s_refptr<Symbol> a, s_refptr<Symbol> b, s_str& e) -> s_refptr<Symbol> {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(a), sb = s_refptr<Scalar>::StaticCast(b);
        return s_refptr<Symbol>(new Scalar("",
            sa->vr*sb->vr - sa->vi*sb->vi,
            sa->vr*sb->vi + sa->vi*sb->vr
        ));
    });
    mState.RegisterBinaryOperator<Scalar,Scalar>("/", [] (State& s, s_refptr<Symbol> a, s_refptr<Symbol> b, s_str& e) -> s_refptr<Symbol> {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(a), sb = s_refptr<Scalar>::StaticCast(b);
        s_double invb = 1.0/(sb->vr*sb->vr + sb->vi*sb->vi);
        return s_refptr<Symbol>(new Scalar("",
            invb*(sa->vr*sb->vr + sa->vi*sb->vi),
            invb*(sa->vr*sb->vi - sa->vi*sb->vr)
        ));
    });
    mState.RegisterBinaryOperator<Scalar,Scalar>("+", [] (State& s, s_refptr<Symbol> a, s_refptr<Symbol> b, s_str& e) -> s_refptr<Symbol> {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(a), sb = s_refptr<Scalar>::StaticCast(b);
        return s_refptr<Symbol>(new Scalar("", sa->vr + sb->vr, sa->vi + sb->vi));
    });
    mState.RegisterBinaryOperator<Scalar,Scalar>("-", [] (State& s, s_refptr<Symbol> a, s_refptr<Symbol> b, s_str& e) -> s_refptr<Symbol> {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(a), sb = s_refptr<Scalar>::StaticCast(b);
        return s_refptr<Symbol>(new Scalar("", sa->vr - sb->vr, sa->vi - sb->vi));
    });
    mState.RegisterBinaryOperator<Scalar,Scalar>("^", [] (State& s, s_refptr<Symbol> a, s_refptr<Symbol> b, s_str& e) -> s_refptr<Symbol> {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(a), sb = s_refptr<Scalar>::StaticCast(b);
        if (sa->vr < 0.0)
        {
            if (!e.IsEmpty()) e.PushBack('\n');
            e += "\tError : The power operator '^' is only defined for real numbers as right operand";
            return nullptr;
        }
        s_double r = pow(sa->vr*sa->vr + sa->vi*sa->vi, 0.5*sb->vr);
        s_double theta = atan2(sa->vi, sa->vr)*sb->vr;
        return s_refptr<Symbol>(new Scalar("", r*cos(theta), r*sin(theta)));
    });
    mState.RegisterBinaryOperator<Scalar,Scalar>("%", [] (State& s, s_refptr<Symbol> a, s_refptr<Symbol> b, s_str& e) -> s_refptr<Symbol> {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(a), sb = s_refptr<Scalar>::StaticCast(b);
        if (sa->vi != 0.0 || sb->vi != 0.0)
        {
            if (!e.IsEmpty()) e.PushBack('\n');
            e += "\tError : The modulo operator '%' is only defined for real numbers";
            return nullptr;
        }
        return s_refptr<Symbol>(new Scalar("", (long)sa->vr.Get() % (long)sb->vr.Get()));
    });
    mState.RegisterRightUnaryOperator<Scalar>("+", [] (State& s, s_refptr<Symbol> a, s_str& e) -> s_refptr<Symbol> {
        return a->Clone();
    });
    mState.RegisterRightUnaryOperator<Scalar>("-", [] (State& s, s_refptr<Symbol> a, s_str& e) -> s_refptr<Symbol> {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(a);
        return s_refptr<Symbol>(new Scalar("", -sa->vr, -sa->vi));
    });

    // Functions
    mState.RegisterFunction("rand", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        res = {s_refptr<Symbol>(new Scalar("", s_double::Random()))};
        return true;
    });
    mState.RegisterFunction<Scalar,Scalar>("rand", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]), sb = s_refptr<Scalar>::StaticCast(args[1]);
        res = {s_refptr<Symbol>(new Scalar("", s_double::Random(sa->vr, sb->vr), s_double::Random(sa->vi, sb->vi)))};
        return true;
    });
    mState.RegisterFunction<Scalar,Scalar>("min", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]), sb = s_refptr<Scalar>::StaticCast(args[1]);
        res = {s_refptr<Symbol>(new Scalar("", s_double::Min(sa->vr, sb->vr), s_double::Min(sa->vi, sb->vi)))};
        return true;
    });
    mState.RegisterFunction<Scalar,Scalar>("max", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]), sb = s_refptr<Scalar>::StaticCast(args[1]);
        res = {s_refptr<Symbol>(new Scalar("", s_double::Max(sa->vr, sb->vr), s_double::Max(sa->vi, sb->vi)))};
        return true;
    });
    mState.RegisterFunction<Scalar>("round", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        res = {s_refptr<Symbol>(new Scalar("", s_double::Round(sa->vr), s_double::Round(sa->vi)))};
        return true;
    });
    mState.RegisterFunction<Scalar>("round_up", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        res = {s_refptr<Symbol>(new Scalar("", s_double::RoundUp(sa->vr), s_double::RoundUp(sa->vi)))};
        return true;
    });
    mState.RegisterFunction<Scalar>("round_down", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        res = {s_refptr<Symbol>(new Scalar("", s_double::RoundDown(sa->vr), s_double::RoundDown(sa->vi)))};
        return true;
    });
    mState.RegisterFunction<Scalar>({"square_root", "sqrt"}, [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        s_double r = pow(sa->vr*sa->vr + sa->vi*sa->vi, s_double(0.25));
        s_double theta = atan2(sa->vi, sa->vr)*0.5;
        res = {s_refptr<Symbol>(new Scalar("", r*cos(theta), r*sin(theta)))};
        return true;
    });
    mState.RegisterFunction<Scalar>("exp", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        s_double r = exp(sa->vr);
        res = {s_refptr<Symbol>(new Scalar("", r*::cos(sa->vi.Get()), r*::sin(sa->vi.Get())))};
        return true;
    });
    mState.RegisterFunction<Scalar>("log", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi != 0.0)
        {
            if (!e.IsEmpty()) e.PushBack('\n');
            e += "\tError : 'log' is only defined for real numbers";
            return false;
        }
        res = {s_refptr<Symbol>(new Scalar("", log(sa->vr.Get())))};
        return true;
    });
    mState.RegisterFunction<Scalar>("cos", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi == 0.0)
            res = {s_refptr<Symbol>(new Scalar("", ::cos(sa->vr.Get())))};
        else
        {
            res = {s_refptr<Symbol>(new Scalar("",
                ::cos(sa->vr.Get())*::cosh(sa->vi.Get()),
                -::sin(sa->vr.Get())*::sinh(sa->vi.Get())
            ))};
        }
        return true;
    });
    mState.RegisterFunction<Scalar>({"cosh", "ch"}, [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi == 0.0)
            res = {s_refptr<Symbol>(new Scalar("", ::cosh(sa->vr.Get())))};
        else
        {
            res = {s_refptr<Symbol>(new Scalar("",
                ::cos(sa->vi.Get())*::cosh(sa->vr.Get()),
                ::sin(sa->vi.Get())*::sinh(sa->vr.Get())
            ))};
        }
        return true;
    });
    mState.RegisterFunction<Scalar>("acos", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi != 0.0)
        {
            if (!e.IsEmpty()) e.PushBack('\n');
            e += "\tError : 'acos' is only defined for real numbers";
            return false;
        }
        res = {s_refptr<Symbol>(new Scalar("", ::acos(sa->vr.Get())))};
        return true;
    });
    mState.RegisterFunction<Scalar>("sin", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi == 0.0)
            res = {s_refptr<Symbol>(new Scalar("", ::sin(sa->vr.Get())))};
        else
        {
            res = {s_refptr<Symbol>(new Scalar("",
                ::sin(sa->vr.Get())*::cosh(sa->vi.Get()),
                ::cos(sa->vr.Get())*::sinh(sa->vi.Get())
            ))};
        }
        return true;
    });
    mState.RegisterFunction<Scalar>({"sinh", "sh"}, [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi == 0.0)
            res = {s_refptr<Symbol>(new Scalar("", ::sinh(sa->vr.Get())))};
        else
        {
            res = {s_refptr<Symbol>(new Scalar("",
                ::cos(sa->vi.Get())*::sinh(sa->vr.Get()),
                ::sin(sa->vi.Get())*::cosh(sa->vr.Get())
            ))};
        }
        return true;
    });
    mState.RegisterFunction<Scalar>("asin", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi != 0.0)
        {
            if (!e.IsEmpty()) e.PushBack('\n');
            e += "\tError : 'asin' is only defined for real numbers";
            return false;
        }
        res = {s_refptr<Symbol>(new Scalar("", ::asin(sa->vr.Get())))};
        return true;
    });
    mState.RegisterFunction<Scalar>("tan", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi != 0.0)
            res = {s_refptr<Symbol>(new Scalar("", ::tan(sa->vr.Get())))};
        else
        {
            s_double r = 2.0/(::cos(2.0*sa->vr.Get()) + ::cosh(2.0*sa->vi.Get()));
            res = {s_refptr<Symbol>(new Scalar("",
                r*::sin(sa->vr.Get())*::cosh(sa->vi.Get()),
                r*::cos(sa->vr.Get())*::sinh(sa->vi.Get())
            ))};
        }
        return true;
    });
    mState.RegisterFunction<Scalar>({"tanh", "th"}, [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi != 0.0)
            res = {s_refptr<Symbol>(new Scalar("", ::tanh(sa->vr.Get())))};
        else
        {
            s_double r = 2.0/(::cos(2.0*sa->vi.Get()) + ::cosh(2.0*sa->vr.Get()));
            res = {s_refptr<Symbol>(new Scalar("",
                r*::cos(sa->vi.Get())*::sinh(sa->vr.Get()),
                r*::sin(sa->vi.Get())*::cosh(sa->vr.Get())
            ))};
        }
        return true;
    });
    mState.RegisterFunction<Scalar>("atan", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        if (sa->vi != 0.0)
        {
            if (!e.IsEmpty()) e.PushBack('\n');
            e += "\tError : 'atan' is only defined for real numbers";
            return false;
        }
        res = {s_refptr<Symbol>(new Scalar("", ::atan(sa->vr.Get())))};
        return true;
    });
    mState.RegisterFunction<Scalar,Scalar>("atan", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]), sb = s_refptr<Scalar>::StaticCast(args[1]);
        if (sa->vi != 0.0 || sb->vi != 0.0)
        {
            if (!e.IsEmpty()) e.PushBack('\n');
            e += "\tError : 'atan' is only defined for real numbers";
            return false;
        }
        res = {s_refptr<Symbol>(new Scalar("", ::atan2(sa->vr.Get(), sb->vr.Get())))};
        return true;
    });
    mState.RegisterFunction<Scalar>("abs", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        res = {s_refptr<Symbol>(new Scalar("", pow(sa->vr*sa->vr + sa->vi*sa->vi, s_double(0.5))))};
        return true;
    });
    mState.RegisterFunction<Scalar>({"conjugate", "conj"}, [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        res = {s_refptr<Symbol>(new Scalar("", sa->vr, -sa->vi))};
        return true;
    });
    mState.RegisterFunction<Scalar>({"real", "re"}, [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        res = {s_refptr<Symbol>(new Scalar("", sa->vr))};
        return true;
    });
    mState.RegisterFunction<Scalar>({"imaginary", "im"}, [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        s_refptr<Scalar> sa = s_refptr<Scalar>::StaticCast(args[0]);
        res = {s_refptr<Symbol>(new Scalar("", sa->vi))};
        return true;
    });
    mState.RegisterFunction<>("pi", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        res = {s_refptr<Symbol>(new Scalar("", M_PI))};
        return true;
    });
    mState.RegisterFunction<>("i", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        res = {s_refptr<Symbol>(new Scalar("", 0.0, 1.0))};
        return true;
    });
    mState.RegisterFunction<>("machine_epsilon", [] (State& s, const s_ctnr<s_refptr<Symbol>>& args, s_ctnr<s_refptr<Symbol>>& res, s_str& e) -> s_bool {
        res = {s_refptr<Symbol>(new Scalar("", epsilon))};
        return true;
    });
}
